// Copyright 1994-2000 by Jon Dart.  All Rights Reserved.

#ifndef _BOARD_H
#define _BOARD_H

#include "types.h"
#include "square.h"
#include "piece.h"
#include "bitmap.h"
#include "material.h"
#include "move.h"
#include <iostream>

class Board_State;

class Board
{
   // This class encapsulates the chess board, and provides functions
   // for making and undoing moves.          

public: 

   friend class BoardIO;
   friend class Scoring;
   
   enum CheckStatusType { InCheck, NotInCheck, Unknown };

   enum CastleType { CanCastleEitherSide,
                     CanCastleKSide,
                     CanCastleQSide,
                     CastledKSide,
                     CastledQSide,
                     CantCastleEitherSide};
          
   // creates a board set up in its initial position, White to move.
   Board();
           
  ~Board();
  
   // Copy constructor and assignment operator  
   Board(const Board &);
   Board &operator = (const Board &);

   // resets board to initial position
   void Reset();
           
#ifdef _DEBUG
   const Piece &operator[]( const Square sq ) const;
#else
   const Piece &operator[]( const Square sq ) const
   {
       return my_Contents[sq];
   }
#endif

   void setContents( const Piece &p, const Square sq )
   {
       my_Contents[sq] = p;
   }
   
   void set_secondary_vars();
   // update all fields of the board, based on the piece positions.

   CastleType CastleStatus( const ColorType side ) const
   {
       return my_CastleStatus[side];        
   }
   
   void setCastleStatus( const CastleType t, const ColorType side )
   {
       my_CastleStatus[side] = t;
   }

   int castled(const ColorType side) const
   {
      return my_CastleStatus[side] == CastledKSide ||
             my_CastleStatus[side] == CastledQSide;
   }
  
   ColorType Side() const
   // side to move         
   {
       return side;       
   }
   
   ColorType OppositeSide() const
   {
       return OppositeColor(side);
   }
   
   void setSide( const ColorType side )
   {
       this->side = side;
   }
   
   const Material &getMaterial( const ColorType side ) const
   {
       return my_Material[side];
   }
   
   const Bitmap &PawnBits( const ColorType side ) const
   {
       return pawn_bits[side];
   }
   
   const Square EnPassantSq()  const
   {
     return my_EnPassantSq;
   }
           
   CheckStatusType CheckStatus( ) const;
   
   CheckStatusType CheckStatus(Move lastMove) const;

   // returns InCheck if the move could cause check to the opponent
   // after being made (call before the move is made).
   CheckStatusType wouldCheck(Move lastMove) const;
   
   void set_check_status(CheckStatusType s)
   {
       my_CheckStatus = s;
   }

   const Square &KingPos( const ColorType side ) const
   {
        return my_KingPos[side];           
   }
           
   hash_t HashCode() const
   {
       return hashCode;
   }
   
   hash_t pawn_hash() const
   {
       return m_pawn_hash;
   }
   
   int operator == ( const Board &ABoard )
   {
       return hashCode == ABoard.HashCode();
   }
   
    // fill "squares" with a list of squares between sq1 and sq2 (in
   // a straight line), returns the # of squares.
   int Between( const Square &sq1, const Square &sq2, 
                Square *squares) const;
   
   // returns "true" if there is a clear path between sq1 and sq2
   int Clear( const Square &sq1, const Square &sq2 ) const;

   // makes a move
   void DoMove(const Move &m);

   // makes a null move
   void DoNull();

   // undoes a previous move.
   void UndoMove( const Move rmove, const Board_State &stat );
   
   // makes a move during the qsearch. The hashcode is not updated.
   void DoMoveQ(const Move &m);
   
   // Return "TRUE" if any piece of color "side" attacks "sq".
   int any_attacks(const Square sq, const ColorType side) const;

   // Return "TRUE" if any piece of color "side" attacks "sq".
   int any_attacks(const Square sq, const ColorType side, Bitmap &source) const;

   // Return a bit vector of all pieces of color "side" that attack "sq".
   Bitmap calc_attacks(const Square sq, const ColorType side) const;
   
   // Return the least-valued piece of color "side" out of the bitmap.
   Bitmap min_attacker(const Bitmap &, const ColorType side) const;
   
   // Return all squares attacked by pawns of color "side".
   Bitmap all_pawn_attacks(const ColorType side) const;
   
   int rep_count() const;

   friend istream & operator >> (istream &i, Board &board);
   friend ostream & operator << (ostream &o, const Board &board);

#if !defined(_WIN32) && defined(DEBUG_ATTACKS)
   // for debugging :
   void dump_attacks()
   {
       my_attacks.dump_attacks();           
   }
#endif   
   
   private:
           
   Piece my_Contents[64];
   Square my_KingPos[2];
   Material my_Material[2];

public:
   ColorType side;
   // caution: Board_State relies on the order of the following 8 member variables:
   hash_t hashCode;
   hash_t m_pawn_hash;
   CastleType my_CastleStatus[2];
   Square my_EnPassantSq;
   CheckStatusType my_CheckStatus;
   int move_count; // since last capture, pawn push or castle

   hash_t *rep_list; //  move history for repetition detection

   hash_t *rep_list_head; // head of history list
   Bitmap pawn_bits[2];
   Bitmap knight_bits[2];
   Bitmap bishop_bits[2];
   Bitmap rook_bits[2];
   Bitmap rooksqueens[2];
   Bitmap bishopsqueens[2];
   Bitmap queen_bits[2];
   Bitmap occupied[2];
   Bitmap all_occupied;
   Bitmap occupiedR90;
   Bitmap occupiedDa1;
   Bitmap occupiedDa8;

private:
   CheckStatusType checkDiscovered(
       Square startSquare, Square kp, int direction) const;

   void UpdateCastleStatus( const ColorType side, 
                            const Square sq );
   void undo_castling(const Square &kp,const Square &oldkingsq,
           const Square &newrooksq, const Square &oldrooksq);
           
   void setAll(ColorType side,int n)
   {
      const register masks *masks = &all_masks[n];
      occupied[side].data |= masks->mask;
      occupiedR90.data |= masks->maskR90;
      occupiedDa1.data |= masks->maskDa1;
      occupiedDa8.data |= masks->maskDa8;
   }
   void clearAll(ColorType side,int n)
   {
      const register masks *masks = &all_masks[n];
      occupied[side].data &= masks->not_mask;
      occupiedR90.data &= masks->not_maskR90;
      occupiedDa1.data &= masks->not_maskDa1;
      occupiedDa8.data &= masks->not_maskDa8;
   }
};
  
#include "boardsta.h"

#endif

