// Copyright 1994, 2003 by Jon Dart. All Rights Reserved.
//
//
#include "evasions.h"
#include "bearing.h"
#include "util.h"

static int is_pinned(const Board & board, const Square source,
                      const Square dest)
{
   int pin = 0;
   if (board.any_attacks(source, board.OppositeSide()))
   {
      // It is possible that the piece we plan to move is pinned.
      Piece PinnedByPiece;
      Square PinnedBySquare;
      int dir;
      if (Bearing::Pinned(board, source, PinnedByPiece,
                          PinnedBySquare, dir))
      {
         // This code checks for moves in the direction of the pin,
         // which are ok:
         int dir1 = Util::Abs((int) source - (int) dest);
         int dir2 = Util::Abs(dir);
         if (dir2 == 1 && Rank(source,White) != Rank(dest,White))
	   pin = 1;
         else if (dir2 == RankIncr && File(source) != File(dest))
	   pin = 1;
         else if (dir1 % dir2)
	   pin = 1;
      }
   }
   return pin;
}

unsigned Check_Evasions(const Board &board, Move * moves, int captures_only)
{
   int num_moves = 0;
   unsigned n;
   unsigned i;
   static Square squares[36];
   Square source;
   const Square kp = board.KingPos(board.Side());
   Bitmap king_attacks(board.calc_attacks(kp, board.OppositeSide()));
   ASSERT(!king_attacks.is_clear());
   int num_attacks = 0;
   int fo = king_attacks.first_one();
   num_attacks = king_attacks.bit_count();
   if (num_attacks == 1)
   {
      // try to capture checking piece
      source = (Square)fo;

      ASSERT(source != InvalidSquare);
      Bitmap atcks(board.calc_attacks(source,board.Side()));
      Square sq;
      for (;(sq = atcks.first_one()) != InvalidSquare;atcks.clear(sq))
      {
         PieceType capturingPiece = TypeOfPiece(board[sq]);
         if (capturingPiece == King)
         {
            // We can capture with the king only if the piece
            // checking us is undefended.  But always allow a
            // capture *of* the king - for illegal move detection.
            if (TypeOfPiece(board[source]) == King ||
                !board.any_attacks(source, board.OppositeSide()))
            {
               moves[num_moves++] = CreateMove(board, sq, source, InvalidPiece);
            }
         }
         else
         {
            if (!is_pinned(board, sq, source))
            {
               if (capturingPiece == Pawn &&
                   Rank(source,board.Side()) == 8)
               {
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Queen);
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Rook);
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Knight);
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Bishop);
               }
               else
                  moves[num_moves++] = CreateMove(board,sq, source, InvalidPiece);
            }
         }
      }
      // Bearing::Attack does not return en passant captures, so try
      // this as a special case
      if (board.EnPassantSq() == source)
      {
         Square dest = source + RankIncr * Direction[board.Side()];
         Piece myPawn = MakePiece(Pawn,board.Side());
         if (File(source) != 8 && board[source + 1] == myPawn)
         {
            Piece tmp = board[source];
            Piece &place = (Piece &)board[source];
            place = EmptyPiece; // imagine me gone
            if (!is_pinned(board, source + 1, dest))
               moves[num_moves++] = CreateMove(source + 1, dest, InvalidPiece, Pawn,
                                               Pawn, EnPassant);
            place = tmp;
         }
         if (File(source) != 1 && board[source - 1] == myPawn)
         {
            Piece tmp = board[source];
            Piece &place = (Piece &)board[source];
            place = EmptyPiece; // imagine me gone
            if (!is_pinned(board, source - 1, dest))
               moves[num_moves++] = CreateMove(source - 1, dest, InvalidPiece, Pawn,
                                               Pawn, EnPassant);
            place = tmp;
         }
      }
      if (captures_only)
         return num_moves;
      // try to interpose a piece
      if (TypeOfPiece(board[source]) != Knight &&
          TypeOfPiece(board[source]) != Pawn)
      {
         static Square btwn_squares[8];
         unsigned nbsq = board.Between(source, kp, btwn_squares);
         ASSERT(nbsq<8);
         for (i = 0; i < nbsq; i++)
         {
            n = Bearing::Attack(board, btwn_squares[i],
                                board.Side(), squares);
            for (unsigned j = 0; j < n; j++)
            {
               Piece p = board[squares[j]];
               if (TypeOfPiece(p) != King &&
                   !is_pinned(board, squares[j], btwn_squares[i]))
               {
                  Square dest = btwn_squares[i];
                  if (TypeOfPiece(p) == Pawn &&
                      Rank(dest,board.Side()) == 8)
                  {
                     // interposition is a promotion
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Queen);
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Rook);
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Knight);
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Bishop);
                  }
                  else
                  {
                     moves[num_moves++] = CreateMove(board, squares[j], dest, InvalidPiece);
                  }
               }
            }
         }
      }
   }
   if (captures_only)
      return num_moves;
   // generate evasive moves
   n = Bearing::BearSq(board, kp, squares);
   for (i = 0; i < n; i++)
   {
      if (!board.any_attacks(squares[i], board.OppositeSide()))
         if (num_attacks > 1 || squares[i] != source)
         {
            // We need to do some extra checking, since the board
            // info on attacks reflects the state before the move,
            // and we need to be sure that the destination square
            // is not attacked after the move.
	   int illegal = 0;
            Bitmap it(king_attacks);
            Square atck_sq;
            while ((atck_sq = it.first_one()) != InvalidSquare && !illegal)
            {
               Piece attacker = board[(Square)(atck_sq)];
               if (sliding(attacker))
               {
                  int dir = Bearing::Directions[atck_sq][squares[i]];
                  // check for movement in the direction of the
                  // attacker:
                  if (dir != 0 && (kp + dir == squares[i]))
		    illegal = 1;
               }
               it.clear(atck_sq);
            }
            if (!illegal)
            {
               moves[num_moves++] = CreateMove(board, kp, squares[i], InvalidPiece);
            }
         }
   }
   return num_moves;
}

