// Copyright 1994-1998 by Jon Dart.  All Rights Reserved.

#include "moveord.h"
#include "movegen.h"
#include "scoring.h"
#include "constant.h"
#include "attacke.h"
#include "bearing.h"
#include "util.h"
#include "globals.h"
#include "config.h"
#include "boardsta.h"

#include <iostream>

extern Move Killers1[Constants::MaxPly];
extern Move Killers2[Constants::MaxPly];

inline void swap( Move moves[], const int i, const int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i]; 
   moves[i] = tmp;
}

// move scoring routine, used only at ply 0
int Move_Ordering::score(Board &board, const Move &move )
{
   int val = 0;
   Square dest_square = DestSquare(move);
   Piece capture = board[dest_square];
   if (!IsEmptyPiece(capture))
   {
      if (!board.any_attacks(dest_square, board.OppositeSide()))
         // attack on undefended piece
      {
         val += PieceValue(capture);
      }
      else if (TypeOfMove(move) == Normal)
      {
         int est = attack_estimate(board,move);
         if (est > 0)
            val += est;
         else // losing capture
            val += 10 + (est/16);
      }
      else if (TypeOfMove(move) == Promotion)
      {
         val += PieceValue(PromoteTo(move)) - PieceValue(Pawn);
      }
      else // en-passant?
         val += 32;
   }
   if (EqualMove(move,Killers1[0]) || EqualMove(move,Killers2[0]))
      val += 30;
   Board_State stat(board);
   board.DoMove(move);
   if (board.any_attacks(board.KingPos(board.OppositeSide()),board.Side()))
   {
      // move into check, illegal
      val = -Constants::BIG;
   }
   else
   {
      val -= Scoring::positional_score(board,-Constants::BIG,Constants::BIG);
      if (board.CheckStatus() == Board::InCheck)
         val += 25;
   }
   board.UndoMove(move,stat);
#ifdef _TRACE
   cout << MoveImage(move) << " score = " << val << endl;
#endif
   return val;
}

int
Move_Ordering::reorder_ply0_moves(const Board & board,
                           Move moves[], const int num_moves,
                           const Move & best_move)
{
   if (num_moves == 0)
      return 0;
   // Move gainful capture moves and killers to the front
   // of the move list.
   static int scores[Move_Generator::MaxMoves];
   int j = 0, i = 0;
#ifdef _TRACE
   int k;
#endif
   for (i = 0; i < num_moves; i++)
   {
      int score = 0;
      const Square dest = DestSquare(moves[i]);
      const Square start = StartSquare(moves[i]);
      const Piece capture = board[dest];

      if (EqualMove(moves[i],best_move))
      {
         scores[j] = Constants::BIG;
#ifdef _TRACE
         cout << "move " << MoveImage(moves[i]) << " (pv)"<< endl;
#endif
         swap(moves,i,j);
         ++j;
      }
      else if (!IsEmptyPiece(capture))
      {
         int start_val;
         //if (TypeOfPiece(board[start]) == King)
         //    start_val = PieceValue(Pawn);
         //else
         start_val = PieceValue(board[start]);
         int capture_val = PieceValue(capture);
         int gain;
         if (capture_val > start_val && PromoteTo(moves[i]) == InvalidPiece)
            gain = capture_val - start_val;
         else
            gain = attack_estimate(board,moves[i]);
#ifdef _TRACE
         cout << "move " << MoveImage(moves[i]) << " est = " << gain << endl;
#endif
         scores[i] = gain;
         if (gain >= -10)
         {
            if (i != j)
            {
               scores[j] = gain;
               swap(moves,i,j);
            }
            ++j;
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         score = PieceValue(PromoteTo(moves[i])) - 64;
#ifdef _TRACE
         cout << "move " << MoveImage(moves[i]) << " est = " << score << endl;
#endif
         if (score > 0)
         {
            scores[j] = score;
            swap(moves,i,j);
            ++j;
         }
      }
      else
      {
         // not a capture move
         if ((StartSquare(moves[i]) == StartSquare(Killers1[0]) &&
              DestSquare(moves[i]) == DestSquare(Killers1[0]))) 
         {
            scores[j] = 5000; // put killers before captures
            swap(moves,i,j);
            ++j;
         }
         else if ((StartSquare(moves[i]) == StartSquare(Killers2[0]) &&
              DestSquare(moves[i]) == DestSquare(Killers2[0]))) {
            scores[j] = 4999; // put killers before captures
            swap(moves,i,j);
            ++j;

         }
      }
   }
   if (j > 1)
      sort_moves(moves,scores,j);
#ifdef _TRACE
   cout << "after ply 0 resort: " ;
   for (k=0; k <j; k++)
      cout << MoveImage(moves[k]) << ' ';
   cout << endl;
#endif
   return num_moves;
}

int Move_Ordering::order_ply0_moves( Board &board, Move moves[],
                                     const int num_moves, const Move &best_move, int &easy)
{
   Scoring::init(board);
   int scores[Move_Generator::MaxMoves];

   int i;
   for (i = 0; i < num_moves; i++)
   {
      if (EqualMove(moves[i],best_move))
         scores[i] = Constants::BIG;
      else
         scores[i] = score(board,moves[i]);
   }
   sort_moves(moves, scores, num_moves);
   int real_num_moves = num_moves;

   easy = 1;
   for (i=num_moves-1; i >=0; --i)
   {
      if (scores[i] == -Constants::BIG)
      {
         --real_num_moves;
      }
      else if (i >0 && scores[0]-100 < scores[i])
	easy = 0;
   }
   return real_num_moves;
}

int
Move_Ordering::order_captures(Board & board,
                              Move moves[], const int num_moves,
                              const int alpha, const int material, int ply)
{
   static int scores[Move_Generator::MaxMoves];
   int n=0;
   int bias = QSEARCH_FORWARD_PRUNE_BIAS;
  
   for (int i = 0; i < num_moves; i++)
   {
      PieceType capture = Capture(moves[i]);
      Square loc = StartSquare(moves[i]);
      if (capture != EmptyPiece)
      {
         int est;
         int start_val = (int)PieceValue(board[loc]);
         if (capture == King) // king capture
         {
            // There is no point sorting the
            // rest of the moves. We return only
            // this move, which signals that
            // the previous move was illegal.
            moves[0] = moves[i];
            return 1;
         }
         else
         {
            est = (int)PieceValue(capture);
         }
         if (PromoteTo(moves[i]) != InvalidPiece)
            est += (int)PieceValue(PromoteTo(moves[i])) - PAWN_VALUE;
         // Futility check: if capturing the piece "for free" will not
         // get our score up to alpha, discard the move
         if (material + bias + est >= alpha)
         {
            est -= start_val;
            if (est > 10) 
            {
               // Always consider capturing a higher-valued piece
               scores[i] = est + (int)capture;
               if (i>n){ 
                  scores[n] = scores[i];
                  moves[n] = moves[i];
               }
               ++n;
            }
            else
            {
               // This capture appears to lose, but we call attack_estimate
               // to make sure
	       est = attack_estimate(board,moves[i]);
#ifdef _TRACE
               indent(ply); cout << "move " << MoveImage(moves[i]) <<
                 " est=" << est << endl;
#endif
               if (bias + material + est >= alpha) {
                  scores[i] = est + (int)capture;
                  if (i>n){ 
                    scores[n] = scores[i];
                    moves[n] = moves[i];
                  }
                  ++n;
               }
            }
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         scores[i] = PieceValue(PromoteTo(moves[i])) - PAWN_VALUE;
         if (scores[i] + PAWN_VALUE + QSEARCH_FORWARD_PRUNE_BIAS + 
             material >= alpha) {
            if (i>n){ 
               scores[n] = scores[i];
               moves[n] = moves[i];
            }
            ++n;
         }
      }
   }
   if (n>1){ 
      // this is generally faster than calling sort_moves, because
      // n is small.
      int swapped = 1;
      while (swapped) {
        swapped = 0;
        for (int i = 0; i < n-1; i++) {
           if (scores[i]<scores[i+1]) {
              ++swapped;
              swap(moves,i,i+1);
              int tmp = scores[i];
              scores[i] = scores[i+1];
              scores[i+1] = tmp;
           }
        }
      }
   }
   return n;
}

