// Copyright 1987-2003 by Jon Dart. All Rights Reserved.

#include "search.h"
#include "scoring.h"
#include "constant.h"
#include "movegen.h"
#include "moveord.h"
#include "util.h"
#include "pinfo.h"
#include "attacke.h"
#include "globals.h"
#include "debug.h"
#include "config.h"
#include "bearing.h"
#ifdef TABLEBASES
#include "tbprobe.h"
#endif
#ifdef _TRACE
   #include <iostream>
#endif

extern Move pv[Constants::MaxPly][Constants::MaxPly];
extern int pv_length[Constants::MaxPly];
static int Time_Check_Interval = 1000;
extern int hash_move_valid( const Board &board, const Move &move);

#ifdef _TRACE
#define UPDATE_PV(move,ply) \
{ \
   pv[ply][ply] = move; \
   for (int j=ply+1;j<ply+1+pv_length[ply+1];j++) \
     pv[ply][j] = pv[ply+1][j]; \
   pv_length[ply] = pv_length[ply+1]+1; \
   indent(ply); \
   cout << "best line[" << ply << "]: ";  \
   for (int j2 = ply; j2 < pv_length[ply]+ply; j2++) \
     cout << MoveImage(pv[ply][j2]) << " "; \
   cout << endl;  \
}
#else
#define UPDATE_PV(move,ply) \
{ \
   pv[ply][ply] = move; \
   for (int j=ply+1;j<ply+1+pv_length[ply+1];j++) \
     pv[ply][j] = pv[ply+1][j]; \
   pv_length[ply] = pv_length[ply+1]+1; \
}
#endif

static const int Illegal = -Constants::BIG - 1000;

inline int Search::NullOk(const Board &board,int in_check,int ply,
int depth) {
  if (null_depth && !in_check &&
       !IsNull(last_move[ply-1]) && !board.getMaterial(Black).no_pieces() &&
       !board.getMaterial(White).no_pieces() ) {
     int n = board.getMaterial(board.Side()).has_queen() ? 2 : 3;
     return
         (board.getMaterial(board.Side()).majorPieceCount() >= n);
  }
  else return 0;
}

int Search::
move_search(Board & board, int alpha, int beta,
            const int ply, int depth, 
            int &terminate)
{
   // recursive function, implements alpha/beta search.  We use
   // the PVS (principal variation search) variant of alpha/beta.

   if (time_check_counter >= Time_Check_Interval)
   {
      time_check_counter = 0;
      terminate = check_time(board,ply,0);
   }
   ASSERT(ply < Constants::MaxPly);
   time_check_counter++;
   num_nodes++;
   // moves into check are illegal.
   if (board.any_attacks(board.KingPos(
                                      board.OppositeSide()),board.Side()))
   {
#ifdef _TRACE
      indent(ply); cout << "Illegal move!" << endl;
#endif
      return -Illegal;
   }
   else if (terminate)
   {
      return alpha;
   }

   best[ply] = NullMove; // move we will return

   int in_check = 0;
   int forced = 0;
   int deepFlag = 0;
   int threat = 0;
   Move hash_move = NullMove;
   Position_Info *hash_entry = NULL;      
   
   int rep_count;
#ifdef FUTILITY_PRUNE
   int est;
#endif
   int score = alpha;
//   int avoid_null = 0;

   if (is_draw(board,rep_count,ply))
   {
#ifdef _TRACE
      indent(ply); cout << "draw!" << endl;
#endif
      score = draw_score(board);
      if (score > alpha && score < beta)
      {
          pv[ply][ply] = NullMove;
          pv_length[ply] = 0;
      }
      return score;
   }
   else if (ply >= Constants::MaxPly-1)
   {
      return evalu8(board,alpha,beta,ply);
   }
   if (use_hash_table)
   {
      // Search the hash table to see if we have hit this
      // position before.
      //
      hash_entry = searchHash(board,rep_count);
      hash_searches++;
      if (!hash_entry)	{
	  // Try the permanent hash table
	  if (position_book && options.learning.position_learning)
	    hash_entry = position_book->lookup(board,rep_count);
#ifdef _TRACE
	  if (hash_entry)
	    cout << "position book hash hit" << endl;
#endif 
      }
      if (hash_entry)
      {
         hash_move = hash_entry->best_move(board);
#ifdef _DEBUG
         // Found the position in the hash table.
         if (!hash_move_valid(board,hash_move))
         {
#ifdef _TRACE
            indent(ply);
            cout << "Warning: hash move invalid" << endl;
#endif              
            hash_entry = NULL;
            hash_move = NullMove;
         }
      }
      if (hash_entry)
      {
#endif
         hash_hits++;
         forced = hash_entry->forced();
         board.set_check_status(hash_entry->in_check() ?
                               Board::InCheck : Board::NotInCheck);
//	     avoid_null = hash_entry->type() == Position_Info::UpperBound &&
//	           (hash_entry->depth() >= depth-null_depth-DEPTH_INCREMENT &&
//               hash_entry->value() < beta);
         if (hash_entry->depth() >= depth ||
             (hash_entry->type() == Position_Info::Valid &&
              Util::Abs(hash_entry->value()) > Constants::BIG-100))
         {
            int value = hash_entry->value();
            //
            // Whether we can use the hash value or not depends on
            // the flags:
            //
            switch (hash_entry->type())
            {
            case Position_Info::Valid:
            // If this is a mate score, adjust it to reflect the
            // current ply depth.
            //
            if (value > Constants::BIG-100)
            {
               value -= ply;
            }
            else if (value < -Constants::BIG+100)
            {
               value += ply;
            }
            score = value;
#ifdef _TRACE
               indent(ply);
               cout << "hash hit, type = E" <<
               " move = " << MoveImage(hash_move) <<
               " alpha = " << alpha << " beta = " << beta << 
               " value = " << value << endl;
#endif
               if (score > alpha && score < beta)
	       {
                  pv[ply][ply] = hash_move;
#ifdef _TRACE
	indent(ply); cout << "best line[ply][ply] = " << MoveImage(hash_move) << endl;
#endif           
                  pv_length[ply] = 1;
	       }
	       return score;
            case Position_Info::LowerBound:
               if (value >= beta)
               {
#ifdef _TRACE
                  indent(ply);
                  cout << "hash hit, type = L" <<
                  " move = " << MoveImage(hash_move) <<
                  " alpha = " << alpha << " beta = " << beta << 
                  " value = " << value << endl;
#endif
                  return value;
               }
               break;
            case Position_Info::UpperBound:
#ifdef _TRACE
               indent(ply);
               cout << "hash hit, type = U" <<
               " move = " << MoveImage(hash_move) <<
               " alpha = " << alpha << " beta = " << beta << 
               " value = " << value << endl;
#endif
               if (value <= alpha)
               {
                  return value;
               }
               break;
            default:
               break;
            }
         }
      }
   }
#ifdef TABLEBASES
  if (options.search.use_tablebases) {
     const Material &wMat = board.getMaterial(White);
     const Material &bMat = board.getMaterial(Black);
     int tb_hit = 0, tb_score;
     // Like Crafty, we restrict tablebase probes to near the top
     // of the tree, unless a promotion or capture is made.
     if(wMat.piece_count() + bMat.piece_count() <= EGTBMenCount &&
        (ply < 3 || TypeOfMove(last_move[ply-1]) != Normal)) {
       stats->tb_probes++;
       tb_hit = probe_tb(board, tb_score, ply);
     } else tb_hit = 0;
     if (tb_hit) {
#ifdef _TRACE
       indent(ply); cout << "EGTB hit: score " << tb_score << endl;
#endif
        stats->tb_hits++;
        score = tb_score;
        pv[ply][ply] = NullMove;
        pv_length[ply] = 0;
        // insert TB info in hash table. 
        if (score < -Constants::BIG+100)
        {
           score -= ply;
        }
        else if (score > Constants::BIG-100)
        {
           score += ply;
        }
        Position_Info new_pos(board, Constants::MaxPly*DEPTH_INCREMENT,
                              age,
                              Position_Info::Valid, 0,
                              board.CheckStatus() == Board::InCheck,
                              rep_count, score, NullMove);
        switch (storeHash(new_pos.hash_code(),new_pos))
        {
           case 0: hash_inserts++; break;
           case 1: hash_replaces++; break;
           case 2: hash_inserts_failed++; break;
           default:  break;
        }
        return score;
     }
  }
#endif
#ifdef _TRACE
   if (!IsNull(hash_move))
   {
      indent(ply);
      cout << "hash move = " << MoveImage(hash_move) << endl;
   }
#endif
   // At this point we need to know if we are in check or not.
   in_check = checks[ply] = board.CheckStatus(last_move[ply-1]) == Board::InCheck;
#ifdef _TRACE
   indent(ply); cout << "in_check=" << in_check << endl;
#endif
   int fprune=0;
#if defined(RAZORING) || defined(FUTILITY_PRUNE)
   int pruneOk = depth <= RAZOR_DEPTH && !in_check &&
        (board.getMaterial(White).piece_count() +
         board.getMaterial(Black).piece_count() >= 6);
   int mat_score;
   if (pruneOk) mat_score = //Scoring::material_score(board);
      (int)board.getMaterial(board.Side()).value() - 
        (int)board.getMaterial(board.OppositeSide()).value();
#endif
#ifdef RAZORING
   int razor = pruneOk &&
               depth > RAZOR_DEPTH+RAZOR_EXTENSION &&
               (mat_score + RAZOR_MARGIN) < alpha;
#endif
#ifdef FUTILITY_PRUNE
    int fmax,fscore,fpruned_moves=0;
    // futility prune only if not in check and not very low in
    // material (where captures can lead to a tablebase hit or
    // can otherwise change the score dramatically).
   if (pruneOk) {
       fscore = FUTILITY_MARGIN + mat_score;
       if (fscore  < alpha) {
#ifdef _TRACE
         indent(ply); cout << "fprune=1 alpha=" << alpha << " material=" <<
             mat_score << endl;
#endif
         fprune = 1;
         fmax = fscore;
       }
    }
#endif
   int initial_alpha = alpha;
   int hi_bound = beta;
   int extend = 0;
   int best_count = 0, move_count = 0;
   Move best_move = NullMove; // move we will hash
   int num_try = 0; // # of legal moves tried

   Board_State save_state(board);

   // Try to get a fast cutoff using a "null move".  Skip if the side
   // to move is in check or if material balance is low enough that
   // zugzwang is a possibility.
   if (NullOk(board,in_check,ply,depth)) {
      last_move[ply] = NullMove;
      board.DoNull();
#ifdef _TRACE        
      indent(ply);
      cout << "trying " << ply << ". " << "(null)" << endl;
#endif
      int nu_depth = depth-null_depth-DEPTH_INCREMENT;
#ifdef ADAPTIVE_NULL
      if (depth >= 8*DEPTH_INCREMENT ||
          (depth >= 5*DEPTH_INCREMENT && 
           board.getMaterial(board.Side()).majorPieceCount() >= 3))
         nu_depth -= DEPTH_INCREMENT; // R=3 
#endif
      int nscore;
      if (nu_depth >= 0)
         nscore = -move_search(board, -beta, -alpha,
                               ply+1, nu_depth, terminate);
      else
         nscore = -quiesce(board, -beta, -alpha,
                           ply+1, 0);
      //      if (board.getMaterial(board.Side()).material_level()<=3) nscore -= PAWN_VALUE/4;
#ifdef _TRACE
      indent(ply);
      cout << ply << ". " << "(null)" << ' ' << nscore << endl;
#endif
      board.UndoMove(NullMove,save_state);
      if (terminate)
      {
         return alpha;
      }
      else if (nscore >= beta) // cutoff
      {
#ifdef _TRACE
         indent(ply);
         cout << "**CUTOFF**" << endl;
#endif
         stats->num_null_cuts++;
         score = nscore;
         goto search_end;
      }
      if (nscore < 100-Constants::BIG) {
         threat = THREAT_EXTENSION;
         fprune = 0;
      }
   }

   // Use "internal iterative deepening" to get an initial move to try if
   // there is no hash move .. an idea from Crafty.  Note that we do
   // not try this if we have a hash table entry and no hash move, on
   // the theory that we tried it before and didn't get a move.
   int dscore;
   if (IsNull(hash_move) && 
       depth > 2*DEPTH_INCREMENT && 
       (!hash_entry || !hash_entry->deep_flag()))
   {
      int d;
      d = Util::Min(depth-2*DEPTH_INCREMENT,DEPTH_INCREMENT+depth/2);
#ifdef _TRACE
      indent(ply); cout << "== deepening, depth = " << d
      << endl;
#endif
      dscore = move_search(board,alpha,beta,ply,d,terminate);
      if (terminate)
      {
         return alpha;
      }
      if (IsNull(best[ply])) // don't try this again
         deepFlag = 1;
#ifdef _TRACE
      indent(ply); cout << "== deepening done.";
#endif
      hash_move = best[ply];
#ifdef _TRACE
      if (!IsNull(hash_move))
      {
         indent(ply); cout << "  hash_move = "
         << MoveImage(hash_move);
      }
      cout << endl;
#endif        
   }
   {
   Move_Generator mg(board, ply, hash_move);
   forced = forced | (in_check && mg.num_evasions() == 1);
   int try_score;
   //
   // Start searching from the current node, recursively
   //
   int first = 1;
   for (;;) {
      Move move = in_check ? mg.NextEvasion() : mg.NextMove();
      if (IsNull(move)) break;
      if (IsUsed(move)) continue;
      move_count++;
      num_moves++;
#ifdef _TRACE
      indent(ply);
      cout << "trying " << ply << ". " << MoveImage(move);
      if (first) cout << "(pv)";
      cout << endl;
#endif
      last_move[ply] = move;
#ifdef FUTILITY_PRUNE
      int fprune2 = 0;
      if (fprune) {
        est = PieceValue(Capture(move));
        if (TypeOfMove(move) == Promotion)
           est += PieceValue(PromoteTo(move))-PAWN_VALUE;  
        fprune2 = fmax + est < alpha;
      }
#endif
      Board::CheckStatusType in_check_after_move = board.wouldCheck(move);
      // try to forward prune before DoMove, if possible
#ifdef FUTILITY_PRUNE
      if (fprune2 && depth<=DEPTH_INCREMENT && in_check_after_move == Board::NotInCheck &&
       !(PieceMoved(move) == Pawn &&
         Rank(DestSquare(move),board.Side()) >= 7))
      {
#ifdef _TRACE
        indent(ply); cout << "fwd pruned." << endl;
#endif
        ++fpruned_moves;
        continue;
      }
#endif
      board.DoMove(move);
      if (in_check_after_move == Board::Unknown)  {
         in_check_after_move = board.CheckStatus();
      }
      else {
         ASSERT(board.CheckStatus() == in_check_after_move);
         board.set_check_status(in_check_after_move);
      }
      extend = calc_extensions(board,ply,depth,in_check_after_move == Board::InCheck,
          forced,threat);
#ifdef RAZORING
      if (razor && TypeOfMove(move) != Promotion && 
         (Capture(move) == EmptyPiece) && extend == 0) {
          extend += RAZOR_EXTENSION;
          ++stats->num_razored;
#ifdef _TRACE
          indent(ply); cout << "razoring, new depth = " << depth+RAZOR_EXTENSION << endl;
#endif
      }
#endif
#ifdef FUTILITY_PRUNE
      if (fprune2 && depth+extend<=DEPTH_INCREMENT && in_check_after_move == Board::NotInCheck &&
       !(PieceMoved(move) == Pawn &&
         Rank(DestSquare(move),board.OppositeSide()) >= 7))
      {
#ifdef _TRACE
        indent(ply); cout << "fwd pruned." << endl;
#endif
        board.UndoMove(move,save_state);
        ++fpruned_moves;
        continue;
      }
#endif    
      if (depth+extend-DEPTH_INCREMENT >= 0)
      {
         try_score = -move_search(board, -hi_bound, -alpha,
                                  ply+1, depth+extend-DEPTH_INCREMENT,
                                  terminate);
      }
      else
      {
         try_score = -quiesce(board, -hi_bound, -alpha,
                              ply+1, 0);
      }
      if (try_score != Illegal && !first && (try_score > alpha) && (try_score < beta))
      {
         // We failed to get a cutoff and must re-search
#ifdef _TRACE
         indent(ply); cout << "window = [" << alpha << "," << beta
         << "]" << endl;
         indent(ply); cout << "score = " << try_score << " - no cutoff, researching .." << endl;
#endif
         if (depth+extend-DEPTH_INCREMENT >= 0)
            try_score=-move_search(board,
                                   -beta,-alpha,ply+1,depth+extend-DEPTH_INCREMENT,terminate);
         else
            try_score=-quiesce(board,
                               -beta,-alpha,ply+1,0);
#ifdef _TRACE
         indent(ply);
         cout << ply << ". " << MoveImage(move) << ' ';
         cout << try_score;
         cout << endl;
#endif
      }
      board.UndoMove(move,save_state);
      if (terminate)
      {
         return alpha;
      }
#ifdef _TRACE
      indent(ply);
      cout << ply << ". " << MoveImage(move) << ' ';
      cout << try_score;
      if (first) cout << " (pv)";
      cout << endl;
#endif
      if (try_score != Illegal)
      {
         first = 0;
         num_try++;
         if (try_score > alpha)
         {
            best_count = move_count-1;
            score = try_score;
            best_move = move;
            if (try_score >= beta)
            {
              goto beta_cutoff;
            }
            best[ply] = move;
            alpha = try_score;
         }
         // check for mate score - no point searching further if
         // there is a mate
         if (try_score >= Constants::BIG-1-ply) goto beta_cutoff;
         // zero-width window for the rest of the moves
         hi_bound=alpha+1;
      }
   }
   beta_cutoff: ;
#ifdef _TRACE
   if (score >= beta)
   {
      indent(ply);
      cout << "**CUTOFF**" << endl;
   }
#endif
   if (num_try == 0 
#ifdef FUTILITY_PRUNE
	   && fpruned_moves == 0
#endif
	   )
   {
      // no moves were tried
      if (in_check)
      {
         if (mg.num_evasions() == 0)        // mate
         {
#ifdef _TRACE
            indent(ply); cout << "mate" << endl;
#endif               
            score = -(Constants::BIG - ply);
            if (score > alpha) alpha = score;
            goto update_pv;
         }
         else
         {
            // We generated some moves, and some were legal,
            // but we skipped them all.
            score = evalu8(board,alpha,beta,ply);
            if (score > alpha) alpha = score;
         }
      }
      else    // stalemate
      {
#ifdef _TRACE
         indent(ply); cout << "stalemate!" << endl;
#endif
         score = draw_score(board);
         if (score > alpha) alpha = score;
         goto update_pv; // skip the hash insert
      }
   }
   else if (num_try > 0)
   {
      if (score >= beta)
      {
         // If the last move caused cutoff, and was not a capture,
         // save it as a "killer" move:
         if (Capture(last_move[ply]) == EmptyPiece &&
             PromoteTo(last_move[ply]) == InvalidPiece)
            Move_Generator::set_killer(board,last_move[ply], ply, depth);
         Move_Generator::update_history(board,last_move[ply],depth);
      }
      stats->move_order_count++;
      if (best_count<4) {
        stats->move_order[best_count]++;
      }
   }
   }

   search_end:
   // don't insert into the hash table if we are terminating - we may
   // not have an accurate score.
   if (use_hash_table && !terminate)
   {
      if (best_move == NullMove) best_move = hash_move;
      // store the position in the hash table, if there's room
      int value = score;
      Position_Info::ValueType val_type;
      // Adjust mate scores to reflect current ply. But only
      // if the score is in bounds.
      if (value > initial_alpha && value < beta)
      {
         if (value < -Constants::BIG+100)
         {
            value -= ply;
         }
         else if (value > Constants::BIG-100)
         {
            value += ply;
         }
      }
#ifdef _TRACE
      char typeChar;
#endif
      if (value <= initial_alpha)
      {
         val_type = Position_Info::UpperBound;
#ifdef _TRACE
         typeChar = 'U';
#endif
         // We don't have a "best" move, because all moves
         // caused alpha cutoff.  But if there was a hash
         // move or an initial move produced by internal 
         // interative deepening, save it in the hash table
         // so it will be tried first next time around.
         best_move = hash_move;
      }
      else if (value >= beta)
      {
         val_type = Position_Info::LowerBound;
#ifdef _TRACE
         typeChar = 'L';
#endif
      }
      else
      {
         val_type = fpruned_moves ? Position_Info::Invalid : Position_Info::Valid;
#ifdef _TRACE
         typeChar = 'E';
#endif
      }
#ifdef _TRACE
      indent(ply);
      cout << "storing " << "type = " << typeChar << 
      " depth = " << depth << " value = " << value << 
      " move = " << MoveImage(best_move) << endl;
#endif
        Position_Info new_pos(board, depth,
                              age,
                              val_type, forced,
                              in_check,
                              rep_count, value,
                              best_move);
        new_pos.set_deep_flag(deepFlag);
        switch (storeHash(new_pos.hash_code(),new_pos))
        {
           case 0: hash_inserts++; break;
           case 1: hash_replaces++; break;
           case 2: hash_inserts_failed++; break;
           default:  break;
         }
   }
   update_pv:
#ifdef FUTILITY_PRUNE
   stats->futility_pruning += fpruned_moves;
#endif
   if (initial_alpha != alpha && score < beta)
   {
      if (num_try == 0)
         pv_length[ply] = 0;
      else
      {
         UPDATE_PV(best[ply],ply);
         Move_Generator::update_history(board,best[ply],depth);
      }
   }
   ASSERT(score >= -Constants::BIG && score <= Constants::BIG);
   return score;
}

